@@ -1,8 +1,10 @@
+import heapq
 from collections import defaultdict
 
 import networkx as nx
 # from pyscipopt import Model, Pricer, SCIP_RESULT, SCIP_PARAMSETTING, quicksum
 import pyscipopt as scip
+from rs_pricing import Pricer as RustPricer
 
 
 class Label:
@@ -17,42 +19,79 @@ class Label:
             self.visited = set()
         self.last_label = last_label
 
+    def __lt__(self, other):
+        return self.earliest_time < other.earliest_time
+
 
 class Pricer(scip.Pricer):
     """
     Solver for the Resource Constrained Shortest Path Problem, implements a basic Labeling Algorithm.
     """
 
-    def __init__(self, graph, instance, distance_fn=None):
+    def __init__(self, graph, instance, init_added_paths={}, deleted_edges_from_node=set(), distance_fn=None,
+                 strategy="py", verbosity=0):
         super().__init__()
         self.graph = graph
+        self.instance = instance
+        self.deleted_edges_from_node = deleted_edges_from_node
         self.start_depot = instance.depot
         self.end_depot = instance.n_customers + 1
         self.earliest = instance.earliest + [instance.earliest[self.start_depot]]
         self.latest = instance.latest + [instance.latest[self.start_depot]]
         self.demands = instance.demands + [instance.demands[self.start_depot]]
+        self.service_times = instance.service_times + [instance.service_times[self.start_depot]]
         self.capacity = instance.capacity
         self.customers = instance.customers
         self.ncustomers = instance.n_customers
+        self.verbosity = verbosity
         self.init_cons = None
         if distance_fn:
             self.distance_fn = distance_fn
         else:
             self.distance_fn = lambda i, j: self.graph[i][j]['distance']
-        self.time_fn = lambda i, j: self.graph[i][j]['distance']
-        self.added_paths = {}
-        self.elementary = False
+        self.time_fn = lambda i, j: self.graph[i][j]['distance'] + self.service_times[i]
+        self.added_paths = init_added_paths
+
+        assert strategy in ["rust", "py"]
+        self.strategy = strategy
+        if strategy == "rust":
+            self.rust_pricer = self.init_rust_pricer()
+
+    def init_rust_pricer(self) -> RustPricer:
+        drive_times = self.instance.distances
+        for drive_time_list in drive_times:
+            drive_time_list.append(drive_time_list[0])
+
+        neighbors = {n: list(self.graph.neighbors(n)) for n in self.customers}
+        neighbors[self.start_depot] = self.customers
+        neighbors[self.end_depot] = []
+
+        time_windows = [(a, b) for a, b in zip(self.earliest, self.latest)]
+
+        return RustPricer(
+            demands=self.demands,
+            time_windows=time_windows ,
+            service_times=self.service_times,
+            vehicle_capacity=self.capacity,
+            customers=self.customers,
+            start_depot=self.start_depot,
+            end_depot=self.end_depot,
+            drive_time=drive_times,
+            neighbors=neighbors
+        )
 
     def path_from_label(self, label: Label):
         curr = label
         path = []
+        start_times = []
         cost = 0
         while curr is not None:
             path.insert(0, curr.last_node)
+            start_times.insert(0, curr.earliest_time)
             if curr.last_label:
                 cost += self.distance_fn(curr.last_label.last_node, curr.last_node)
             curr = curr.last_label
-        return tuple(path), cost
+        return tuple(path), start_times, cost
 
     def add_path_to_rmp(self, path, cost, solver):
         customer_counts = {}
@@ -78,7 +117,19 @@ class Pricer(scip.Pricer):
                            names=["-".join(map(str, path))],
                            types=['C'])
 
-    def find_path(self, duals, added_paths):
+    def find_path_rust(self, duals, deleted_edges):
+        return self.rust_pricer.find_path(duals, deleted_edges)
+
+    def find_path(self, duals):
+        deleted_edges = self.deleted_edges_from_node[self.model.getCurrentNode().getNumber()]
+        if self.strategy == "py":
+            result = list(self.find_path_py(duals, deleted_edges))
+            result.sort(key=lambda element: str(element[0]))
+            return result
+        elif self.strategy == "rust":
+            return self.find_path_rust(duals, deleted_edges)
+
+    def find_path_py(self, duals, deleted_edges):
         unprocessed = {}
         processed = {}
 
@@ -86,16 +137,19 @@ class Pricer(scip.Pricer):
             unprocessed[i] = set()
             processed[i] = set()
 
-        unprocessed[self.start_depot] = {Label(self.start_depot, 0, 0, 0, None)}
-        unprocessed_count = 1
+        start_label = Label(self.start_depot, 0, 0, 0, None)
+        unprocessed[self.start_depot] = {start_label}
+        label_queue = [(start_label.earliest_time, start_label.cost, start_label)]
+        removed_labels = set()
 
-        while unprocessed_count > 0:
-            label_to_expand = self.choose_label_to_expand(unprocessed)
-            unprocessed_count -= 1
-            # print('***', unprocessed_count)
+        while label_queue:
+            label_to_expand = self.choose_label_to_expand(label_queue, removed_labels)
+            if not label_to_expand: break
             next_node_to_expand = label_to_expand.last_node
+
             for neighbor in nx.neighbors(self.graph, next_node_to_expand):
                 if neighbor in label_to_expand.visited: continue
+                if (next_node_to_expand, neighbor) in deleted_edges: continue
                 demand, last_visited, redcost, earliest_time, visited = self.expand_label(duals, label_to_expand,
                                                                                           neighbor,
                                                                                           next_node_to_expand)
@@ -105,22 +159,34 @@ class Pricer(scip.Pricer):
                 if self.is_feasible(demand, earliest_time, neighbor):
                     dominated = self.dominance_check({new_label}, processed[neighbor] | unprocessed[neighbor])
                     if len(dominated) == 0:
-                        dominated = self.dominance_check(unprocessed[neighbor], {new_label})
-                        unprocessed[neighbor] = (unprocessed[neighbor] - dominated) | {new_label}
-                        unprocessed_count += 1 - len(dominated)
+                        heapq.heappush(label_queue,
+                                       (new_label.earliest_time, new_label.cost, new_label))
+                        if neighbor != self.end_depot:
+                            dominated = self.dominance_check(unprocessed[neighbor], {new_label})
+                            removed_labels |= dominated
+                            unprocessed[neighbor] = (unprocessed[neighbor] - dominated) | {new_label}
             processed[next_node_to_expand].add(label_to_expand)
 
         best_path_label = None
         best_path_redcost = float("inf")
         if processed[self.end_depot]:
             for label in processed[self.end_depot]:
-                # if label.cost < -1e-08:
-                #     yield *self.path_from_label(label), label.cost
+                if label.cost < 1e-6:
+                    yield *self.path_from_label(label), label.cost
                 if label.cost < best_path_redcost:
                     best_path_label = label
                     best_path_redcost = label.cost
-        best_path, best_path_travel_cost = self.path_from_label(best_path_label)
-        yield best_path, best_path_travel_cost, best_path_redcost
+        best_path, start_times, best_path_travel_cost = self.path_from_label(best_path_label)
+        yield best_path, start_times, best_path_travel_cost, best_path_redcost
+
+    def choose_label_to_expand(self, label_heap, removed_labels):
+        while label_heap:
+            *_, label = heapq.heappop(label_heap)
+            if label not in removed_labels:
+                return label
+            else:
+                removed_labels.remove(label)  # this is done for the set size to not blow up
+        return None
 
     def is_feasible(self, demand, earliest_time, neighbor):
         return demand <= self.capacity and earliest_time <= self.latest[neighbor]
@@ -131,7 +197,7 @@ class Pricer(scip.Pricer):
         last_visited = neighbor
         demand = label_to_expand.demand + self.demands[neighbor]
         earliest_time = max(label_to_expand.earliest_time + self.time_fn(next_node_to_expand, neighbor),
-                            self.latest[neighbor])
+                            self.earliest[neighbor])
         visited = set(label_to_expand.visited)
         visited.add(neighbor)
         return demand, last_visited, redcost, earliest_time, visited
@@ -152,62 +218,48 @@ class Pricer(scip.Pricer):
                     break
         return dominated
 
-    @staticmethod
-    def choose_label_to_expand(unprocessed):
-        # todo: choose one with min time
-        # returns first label it sees
-        for customer, labels in unprocessed.items():
-            if len(labels):
-                return labels.pop()
-        return None
-
     def dominates(self, label_a, label_b):
         is_less_or_eq = label_a.cost <= label_b.cost and label_a.demand <= label_b.demand and \
                         label_a.earliest_time <= label_b.earliest_time
         one_is_strictly_less = label_a.cost < label_b.cost or label_a.demand < label_b.demand or \
                                label_a.earliest_time < label_b.earliest_time
-        if self.elementary:
-            visited_less = label_a.visited.issubset(label_b.visited)
-        else:
-            visited_less = True
-        return is_less_or_eq and one_is_strictly_less and visited_less
+        return is_less_or_eq and one_is_strictly_less
 
     def pricerredcost(self, *args, **kwargs):
-        duals = [0]
+        duals = {self.start_depot: 0}
         for i, c in enumerate(self.init_cons):
-            duals.append(self.model.getDualsolLinear(c))
-        duals.append(0)  # end depot dual is the same as the start depot
+            duals[i + 1] = self.model.getDualsolLinear(c)
+        duals[self.end_depot] = 0
 
-        # assert str(paths[0]) not in self.added_paths
         n_added_paths = 0
-        self.elementary = True
+        min_redcost = 0
 
-        while True:
-            something_added = False
-            for path, cost, redcost in self.find_path(duals, self.added_paths):
-                if redcost < -1e-08 and str(path) not in self.added_paths:
-                    something_added = True
-                    n_added_paths += 1
-                    # print(path, redcost)
-                    var = self.model.addVar(name=f"var-{str(path)}", lb=0, ub=float("inf"), obj=cost, vtype="C",
-                                            pricedVar=True)
-                    self.added_paths[str(path)] = var
-                    cust_i_in_path = defaultdict(lambda: 0)
-                    for x in path:
-                        cust_i_in_path[x] += 1
-                    for i, cons in enumerate(self.init_cons):
-                        if cust_i_in_path[i + 1] > 0:
-                            # print(i + 1, cust_i_in_path[i + 1])
-                            self.model.addConsCoeff(cons, var, cust_i_in_path[i + 1])
-            if not something_added and not self.elementary:
-                print("switched to elementary mode")
-                self.elementary = True
-                self.first_found = False
-            else:
-                break
+        for path, start_times, cost, redcost in self.find_path(duals):
+            path_name =str((*path,))
+            if path_name not in self.added_paths:
+                if redcost < min_redcost:
+                    min_redcost = redcost
+                n_added_paths += 1
+                if self.verbosity >= 3:
+                    print(path, start_times, cost, redcost)
 
-
-        return {"result": scip.SCIP_RESULT.SUCCESS}
+                var = self.model.addVar(name=f"{str(path)}", obj=cost, vtype="B",
+                                        pricedVar=True)
+                self.added_paths[path_name] = var
+                cust_i_in_path = defaultdict(lambda: 0)
+                for x in path:
+                    cust_i_in_path[x] += 1
+                for i, cons in enumerate(self.init_cons):
+                    if cust_i_in_path[i + 1] > 0:
+                        # print(i + 1, cust_i_in_path[i + 1])
+                        self.model.addConsCoeff(cons, var, cust_i_in_path[i + 1])
+        if self.verbosity >= 2:
+            print(f"at{self.model.getCurrentNode().getNumber()}, LP obj:", self.model.getLPObjVal())
+            # print("lowerbound", self.model.getLPObjVal() + min_redcost)
+        return {
+            "result": scip.SCIP_RESULT.SUCCESS,
+            "lowerbound": self.model.getLPObjVal() + min_redcost
+        }
 
     def pricerinit(self):
         for i, c in enumerate(self.init_cons):
