@@ -1,5 +1,8 @@
+from functools import lru_cache
+
 from cvrplib.Instance import VRPTW
 import networkx as nx
+from ast import literal_eval as make_tuple
 
 
 def instance_graph(instance: VRPTW):
@@ -24,8 +27,23 @@ def instance_graph(instance: VRPTW):
 
 def minify_instance(instance, only_first):
     instance.customers = instance.customers[:only_first]
-    instance.earliest = instance.earliest[:only_first]
-    instance.latest = instance.latest[:only_first]
-    instance.demands = instance.demands[:only_first]
-    instance.n_customers = only_first - 1
+    instance.earliest = instance.earliest[:only_first + 1]
+    instance.latest = instance.latest[:only_first + 1]
+    instance.demands = instance.demands[:only_first + 1]
+    instance.distances = instance.distances[:only_first + 1]
+    for i in range(only_first + 1):
+        instance.distances[i] = instance.distances[i][:only_first + 1]
+    instance.n_customers = only_first
     return instance
+
+
+def var_to_edges(var):
+    return var_name_to_edges(str(var))
+
+
+@lru_cache(maxsize=None)
+def var_name_to_edges(var_name):
+    var_name = var_name if var_name[0] != "t" else var_name[2:]
+    nodes = list(make_tuple(var_name))
+    nodes[-1] = nodes[0]
+    return set(zip(nodes[:-1], nodes[1:]))
